local wow3 = select(4, GetBuildInfo()) >= 30000

local L = LibStub("AceLocale-3.0"):GetLocale("Utopia")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local SM = LibStub("LibSharedMedia-3.0")
local TalentQuery = LibStub("LibTalentQuery-1.0")
local commPrefix = "Utopia"

local playerClass
local new, del, copy, deepDel, Gradient, UnitFullName, ClassColour, SmoothColour, rotate

local module = LibStub("AceAddon-3.0"):NewAddon("Utopia", "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0", "AceComm-3.0", "AceTimer-3.0")
Utopia = module
module.version = tonumber(strmatch("$Revision: 13 $", "(%d+)")) or 0

local LDB = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("Utopia", {
	type = "data source",
	label = L["TITLE"],
	icon = select(3, GetSpellInfo(47467)),
})

-- LDB Bits
function LDB:Update()
	self.label = Gradient(L["TITLE"], 1, 0.1, 0.1, 1, 1, 0.1)
end

-- OnClick
function LDB.OnClick(self, button)
	if (not IsModifierKeyDown()) then
		if (button == "LeftButton") then
			module:ToggleActive()
		elseif (button == "RightButton") then
			module:OpenConfig()
		end
	end
end

-- OnTooltipShow
function LDB.OnTooltipShow(tt)
	tt:AddLine(Gradient(L["TITLE"].." r"..module.version, 1, 0.1, 0.1, 1, 1, 0.1))
end

module.options = {
	type = 'group',
	order = 15,
	name = L["TITLE"],
	desc = L["Configuration"],
	handler = module,
	args = {
		aceConfig = {
			type = "execute",
			name = L["Standalone Config"],
			desc = L["Open a standalone config window. You might consider installing |cffffff00BetterBlizzOptions|r to make the Blizzard UI options panel resizable."],
			func = function()
				InterfaceOptionsFrame:Hide()
				AceConfigDialog:SetDefaultSize("Utopia", 500, 550)
				AceConfigDialog:Open("Utopia")
			end,
			order = 1,
		},
		enable = {
			type = 'toggle',
			name = L["Enable"],
			desc = L["Enable target debuff status display"],
			get = function() return module.db.char.enable end,
			set = function(info, n) module.db.char.enable = n module:ShowWhen() end,
			order = 5,
		},
		locked = {
			type = 'toggle',
			name = L["Locked"],
			desc = L["Toggle locked position. When unlocked, you can move the display with the bar"],
			get = function() return module.db.char.locked end,
			set = function(info, n) module.db.char.locked = n module.frame.bar:EnableMouse(not module.db.char.locked) end,
			order = 20,
		},
		show = {
			type = 'group',
			name = L["Show When"],
			desc = L["Set rules for when to show Utopia"],
			order = 50,
			args = {
				always = {
					type = 'toggle',
					name = L["Always Show"],
					desc = L["Always show the bar"],
					get = function() return module.db.char.always end,
					set = function(info, n) module.db.char.always = n module:PLAYER_TARGET_CHANGED() end,
					order = 1,
				},
				solo = {
					type = 'toggle',
					name = SOLO,
					desc = SOLO,
					get = function() return module.db.char.showSolo end,
					set = function(info,n) module.db.char.showSolo = n module:ShowWhen() end,
					order = 10,
				},
				party = {
					type = 'toggle',
					name = PARTY,
					desc = PARTY,
					get = function() return module.db.char.showParty end,
					set = function(info,n) module.db.char.showParty = n module:ShowWhen() end,
					order = 20,
				},
				raid = {
					type = 'toggle',
					name = RAID,
					desc = RAID,
					get = function() return module.db.char.showRaid end,
					set = function(info,n) module.db.char.showRaid = n module:ShowWhen() end,
					order = 30,
				},
			},
		},
		display = {
			type = 'group',
			order = 100,
			name = L["Display"],
			desc = L["Display options for the debuff monitor"],
			args = {
				orientation = {
					type = 'select',
					name = L["Orientation"],
					desc = L["Set the orientation for the display"],
					values = {V = L["Vertical"], H = L["Horizontal"]},
					order = 5,
					get = function() return module.db.char.orientation end,
					set = "SetOrientation",
				},
				side = {
					type = 'select',
					name = L["Side"],
					desc = L["Which side to put the bar"],
					values = {B = L["Bottom/Right"], T = L["Top/Left"]},
					order = 10,
					get = function() return module.db.char.side end,
					set = "SetBarOrientation",
				},
				size = {
					type = "range",
					name = L["Icon Size"],
					desc = L["Set the size of the icons"],
					func = timeFunc,
					order = 2,
					get = function() return module.db.char.size end,
					set = "SetIconSize",
					min = 8,
					max = 64,
					step = 1,
					order = 20,
				},
				scale = {
					type = "range",
					name = L["Scale"],
					desc = L["Set the scale of the icons"],
					func = timeFunc,
					order = 2,
					get = function() return module.db.char.scale end,
					set = function(info, n) module.db.char.scale = n module:SetScale() module:SavePosition() end,
					min = 0.2,
					max = 2,
					step = 0.01,
					bigStep = 0.1,
					order = 30,
				},
				texture = {
					type = 'select', dialogControl = 'LSM30_Statusbar',
					name = L["Bar Texture"],
					desc = L["Set the texture for the bar"],
					values = AceGUIWidgetLSMlists.statusbar,
					order = 40,
					get = function() return module.db.char.texture end,
					set = function(v,n)
						module.db.char.texture = n
						local tex = SM:Fetch("statusbar", n)
						module.frame.bar:SetStatusBarTexture(tex)
						module.frame.bar.bg:SetTexture(tex)
					end,
				},
				barwidth = {
					type = "range",
					name = L["Width"],
					desc = L["Set the width of the status bar"],
					func = timeFunc,
					order = 2,
					get = function() return module.db.char.barwidth end,
					set = function(info, n) module.db.char.barwidth = n module:SetBarOrientation() end,
					min = 1,
					max = 20,
					step = 1,
					order = 50,
				},
				headerSize = {
					type = 'header',
					name = L["Colour"],
					order = 100,
				},
				activeColour = {
					type = "color",
					name = L["Active Colour"],
					desc = L["Set the colour for debuffs which your group has applied."],
					hasAlpha = true,
					get = function()
						local c = module.db.char.activeColour
						return c.r, c.g, c.b, c.a
					end,
					set = function(info, r, g, b, a)
						local c = module.db.char.activeColour
						c.r, c.g, c.b, c.a = r, g, b, a
						if (module.frame) then
							module.frame:UpdateBuffs()
						end
					end,
					order = 110,
				},
				inactiveColour = {
					type = "color",
					name = L["Inactive Colour"],
					desc = L["Set the colour for debuffs which your group is able to apply, but has not done so."],
					hasAlpha = true,
					get = function()
						local c = module.db.char.inactiveColour
						return c.r, c.g, c.b, c.a
					end,
					set = function(info, r, g, b, a)
						local c = module.db.char.inactiveColour
						c.r, c.g, c.b, c.a = r, g, b, a
						if (module.frame) then
							module.frame:UpdateBuffs()
						end
					end,
					order = 120,
				},
				unavailableColour = {
					type = "color",
					name = L["Unavailable Colour"],
					desc = L["Set the colour for debuffs which your group is not able to apply."],
					hasAlpha = true,
					get = function()
						local c = module.db.char.unavailableColour
						return c.r, c.g, c.b, c.a
					end,
					set = function(info, r, g, b, a)
						local c = module.db.char.unavailableColour
						c.r, c.g, c.b, c.a = r, g, b, a
						if (module.frame) then
							module.frame:UpdateBuffs()
						end
					end,
					order = 130,
				},
			},
		},
	},
}

AceConfig:RegisterOptionsTable("Utopia", module.options)

-- InitData
function module:InitData()
	del(self.lookup)
	self.lookup = new()

	deepDel(self.classes)
	self.classes = new()

	for name,info in pairs(self.data) do
		info.key = name
		for spell,spellInfo in pairs(info.spells) do
			spellInfo.name = spell
			spellInfo.key = name
			self.lookup[spell] = spellInfo

			local c = self.classes[spellInfo.class]
			if (not c) then
				c = new()
				self.classes[spellInfo.class] = c
			end
			c[name] = spell
		end
	end
end

-- SavePosition
function module:SavePosition()
	self.db.char.pos = {self.frame:GetPoint(1)}
end

-- SavePosition
function module:RestorePosition()
	if (self.db.char.pos) then
		self.frame:SetPoint(unpack(self.db.char.pos))
	else
		self.frame:SetPoint("CENTER", UIParent, "TOP", 0, -150)
	end
end

-- myLevelRank
local function myLevelRank(ranks)
	local lvl = UnitLevel("player")
	local rank
	for i,level in pairs(ranks) do
		if (lvl >= level) then
			rank = i
		end
	end
	return rank
end

-- getAmount
local function getAmount(info, short, debuff)
	local ret = info.amount
	local suffix

	local rank = debuff and debuff.rank
	local stacks = debuff and debuff.stacks

	if (info.amountsPerStack) then
		if (not rank) then
			rank = myLevelRank(info.rankLevels)
		end
		ret = info.amountsPerStack[rank or #info.amountsPerStack]
		assert(ret)
		if (not short) then
			suffix = format(L["(for %d stacks)"], stacks or info.maxStacks)
		end
	elseif (info.amountPerStack) then
		ret = info.amountPerStack
		assert(ret)
		if (not short) then
			suffix = format(L["(for %d stacks)"], stacks or info.maxStacks)
		end
	elseif (info.amounts) then
		if (not rank) then
			rank = myLevelRank(info.rankLevels)
		end
		ret = info.amounts[rank or #info.amounts]
		assert(ret)
	end

	if (info.improvedAmountPerTalentPoint) then
		ret = ret + info.improvedAmountPerTalentPoint * (info.maxTalentPoints or 1)
	elseif (info.improvedPercentagePerTalentPoint) then
		ret = ret * (1 + (info.improvedPercentagePerTalentPoint * (info.maxTalentPoints or 1) / 100))
	elseif (info.amountPerTalentPoint) then
		ret = info.amountPerTalentPoint * (rank or info.maxTalentPoints or 1)
	elseif (info.improvedAmount) then
		ret = info.improvedAmount * (info.maxTalentPoints or 1)
	end

	if (stacks or info.maxStacks) then
		ret = ret * (stacks or info.maxStacks or 1)
		if (not short) then
			suffix = format(L["(for %d stacks)"], stacks or info.maxStacks)
		end
	end

	if (ret and info.amountType) then
		ret = ret..info.amountType
	end

	if (ret and suffix) then
		return format("%s %s", ret, suffix)
	end
	return ret or ""
end

-- InitFrames
function module:InitFrames()
	local frame = CreateFrame("Frame", "UtopiaAnchorFrame", UIParent)
	self.frame = frame
	frame:SetHeight(1)
	frame:SetWidth(1)
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	frame.unit = "target"

	self:SetScale()
	self:RestorePosition()

	frame.icons = {}

	local list = new()
	for key,data in pairs(self.data) do
		tinsert(list, key)
	end
	sort(list)

	local info
	local function sortList(a, b)
		local aa = info.spells[a]
		local bb = info.spells[b]
		aa = getAmount(aa, true)
		bb = getAmount(bb, true)
		
		if (aa == bb) then
			return a < b
		else
			return aa > bb
		end
	end

	local function onEnter(self)
		GameTooltip:SetOwner(self, "LEFT")
		GameTooltip:SetText(self.key, 1, 1, 1)
		info = module.data[self.key]

		local list = new()
		for name,spellInfo in pairs(info.spells) do
			tinsert(list, name)
		end
		sort(list, sortList)

		for i = 1,#list do
			local name = list[i]
			local spellInfo = info.spells[name]
			local key = spellInfo.key
			local data = frame.debuffData and frame.debuffData[key]
			local debuff
			if (data) then
				debuff = data[name]
			end

			local canDo = module.spellPotential and module.spellPotential[name]
			local div = canDo and 1 or 2

			local c = RAID_CLASS_COLORS[spellInfo.class]
			GameTooltip:AddDoubleLine(spellInfo.info and format("%s (%s)", name, spellInfo.info) or name, getAmount(spellInfo, nil, debuff), c.r / div, c.g / div, c.b / div, 1 / div, 1 / div, 1 / div)

			if (canDo) then
      			GameTooltip:AddTexture(debuff and "Interface\\RAIDFRAME\\ReadyCheck-Ready" or "Interface\\RaidFrame\\ReadyCheck-NotReady")
			end
		end
		GameTooltip:Show()
	end
	local function onLeave(self)
		GameTooltip:Hide()
	end

	local function iconOnUpdate(self, elapsed)
		self.angle = self.angle + elapsed * 180			-- Rotate once during fade in/out
		local ULx, ULy, LLx, LLy, URx, URy, LRx, LRy = rotate(self.angle)
		self.shine:SetTexCoord(ULx, ULy, LLx, LLy, URx, URy, LRx, LRy)

		if (self.mode == "in") then
			self.alpha = self.alpha + elapsed
			if (self.alpha > 1) then
				self.alpha = 1
				self.mode = "out"
			end
		else
			self.alpha = self.alpha - elapsed
			if (self.alpha < 0) then
				self.alpha = 0
				self:SetScript("OnUpdate", nil)
			end
		end
		self.shine:SetAlpha(self.alpha)
	end
		
	local function iconActivate(self, debuff)
		local c = module.db.char.activeColour
		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)

		if (not self.shine) then
			self.shine = self:CreateTexture(nil, "OVERLAY")
			self.shine:SetAllPoints()
			self.shine:SetTexture("Spells\\SparkleBlue")
			self.shine:SetBlendMode("ADD")
		end

		if (not self.shine:IsShown()) then
			self.angle = 0
			self.alpha = 0
			self.mode = "in"
			self:SetScript("OnUpdate", iconOnUpdate)
			self.shine:Show()
		end

		if (debuff) then
			local maxStacks
			for spell,data in pairs(debuff) do
				if (data.stacks and (not maxStacks or data.stacks > maxStacks)) then
					maxStacks = data.stacks
				end
			end

			if (maxStacks) then
				self:ShowCount(maxStacks)
			else
				self:ShowCount()
			end
		end
	end

	function iconShowCount(self, count)
		if (count) then
			if (not self.count) then
				self.count = self:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
				self.count:SetAllPoints()
			end
			self.count:SetText(count)
				
		elseif (self.count) then
			self.count:Hide()
		end
	end

	local function iconDeactivate(self)
		local c = module.db.char.inactiveColour
		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)
		if (self.shine) then
			self:SetScript("OnUpdate", nil)
			self.shine:Hide()
		end
	end

	local function iconUnavailable(self)
		local c = module.db.char.unavailableColour
		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)
		if (self.shine) then
			self:SetScript("OnUpdate", nil)
			self.shine:Hide()
		end
	end

	for i = 1,#list do
		local key = list[i]
		local data = self.data[key]

		local icon = CreateFrame("Frame", frame:GetName()..key, frame)
		icon:SetWidth(self.db.char.size)
		icon:SetHeight(self.db.char.size)
		icon:EnableMouse(true)

		icon.key = key

		icon.tex = icon:CreateTexture(nil, "OVERLAY")
		icon.tex:SetAllPoints()
		icon.tex:SetTexture(data.icon)

		if (prev) then
			icon:SetPoint("TOPLEFT", prev, self.db.char.orientation == "H" and "TOPRIGHT" or "BOTTOMLEFT")
		else
			icon:SetPoint("TOPLEFT")
		end

		icon:SetScript("OnEnter", onEnter)
		icon:SetScript("OnLeave", onLeave)

		icon.Activate = iconActivate
		icon.Deactivate = iconDeactivate
		icon.Unavailable = iconUnavailable
		icon.ShowCount = iconShowCount

		tinsert(frame.icons, icon)
		prev = icon
	end
	del(list)

	frame.bar = CreateFrame("StatusBar", frame:GetName().."Bar", frame)
	self:SetBarOrientation()
	
	frame.bar:SetStatusBarTexture(SM:Fetch("statusbar", self.db.char.texture))
	frame.bar:SetStatusBarColor(0, 1, 0)
	frame.bar:SetMinMaxValues(0, #frame.icons)

	frame.bar:RegisterForDrag("LeftButton")
	frame.bar:SetScript("OnDragStart", function(self) module.frame:StartMoving() end)
	frame.bar:SetScript("OnDragStop", function(self) module.frame:StopMovingOrSizing() module:SavePosition() end)

	frame.bar.bg = frame.bar:CreateTexture(nil, "BACKGROUND")
	frame.bar.bg:SetAllPoints()
	frame.bar.bg:SetTexture(SM:Fetch("statusbar", self.db.char.texture))
	frame.bar.bg:SetVertexColor(0, 1, 0, 0.3)

	self:SetFrameSize()
	frame:Hide()

	if (not self.db.char.locked) then
		frame.bar:EnableMouse(true)
	end

	frame:SetScript("OnShow",
		function(self)
			self:UpdateBuffs()
			self:RegisterEvent("UNIT_AURA")

			if (self.unit == "target") then
				self:RegisterEvent("PLAYER_TARGET_CHANGED")
			else
				self:RegisterEvent("PLAYER_FOCUS_CHANGED")
			end
		end
	)
	frame:SetScript("OnHide",
		function(self)
			self:UnregisterEvent("UNIT_AURA")
			self:UnregisterEvent("PLAYER_TARGET_CHANGED")
			self:UnregisterEvent("PLAYER_FOCUS_CHANGED")
		end
	)
	frame:SetScript("OnEvent",
		function(self, event, ...)
			local f = self[event]
			if (f) then
				f(self, ...)
			end
		end
	)

	function frame:UNIT_AURA(unit)
		if (unit == self.unit) then
			self:UpdateBuffs()
		end
	end

	function frame:PLAYER_TARGET_CHANGED()
		self:UpdateBuffs()
	end
	function frame:PLAYER_FOCUS_CHANGED()
		self:UpdateBuffs()
	end

	function frame:UpdateBuffs()
		self:GetDebuffs()
		self:UpdateDisplay()
	end

	function frame:GetDebuffs()
		deepDel(self.debuffData)
		self.debuffData = new()

		if (UnitCanAttack("player", self.unit)) then
			for i = 1,100 do
				local name, rank, buff, count, debuffType, maxDuration, endTime, isMine, isStealable = UnitDebuff(self.unit, i)
				if (not name) then
					break
				end

				local info = module.lookup[name]
				if (info) then
					local data = self.debuffData[info.key]
					if (not data) then
						data = new()
						self.debuffData[info.key] = data
					end

					local n = new()
					n.spellName = name
					n.rank = tonumber(strmatch(rank, "(%d+)"))
					if (count and count ~= 0) then
						n.stacks = count
					end
					n.isMine = isMine

					data[name] = n
				end
			end
		end
	end

	function frame:UpdateDisplay()
		local on, off = 0, 0
		for i,icon in ipairs(self.icons) do
			local key = icon.key

			local c
			local debuff = self.debuffData and self.debuffData[key]
			if (debuff) then
				icon:Activate(debuff)
				on = on + 1
			elseif (module.potential and module.potential[key]) then
				icon:Deactivate()
				off = off + 1
			else
				icon:Unavailable()
			end
		end

		self.bar:SetMinMaxValues(0, on + off)
		self.bar:SetValue(on)

		local r, g, b = SmoothColour(on / (on+off))
		self.bar:SetStatusBarColor(r, g, b)
		self.bar.bg:SetVertexColor(r, g, b, 0.3)
	end

	self.InitFrames = nil
end

-- SetFrameSize
function module:SetFrameSize()
	local frame = self.frame
	if (frame) then
		if (self.db.char.orientation == "H") then
			frame:SetWidth(self.db.char.size * #frame.icons)
			frame:SetHeight(self.db.char.size + self.db.char.barwidth)
		else
			frame:SetWidth(self.db.char.size + self.db.char.barwidth)
			frame:SetHeight(self.db.char.size * #frame.icons)
		end
	end
end

-- SetBarOrientation
function module:SetBarOrientation(info, n)
	if (n) then
		if (n == "B" or n == "T") then
			self.db.char.side = n
		end
	end

	local frame = self.frame
	if (frame) then
		frame.bar:ClearAllPoints()
		if (self.db.char.orientation == "H") then
			if (self.db.char.side == "B") then
				frame.bar:SetPoint("TOPLEFT", frame.icons[1], "BOTTOMLEFT")
				frame.bar:SetPoint("BOTTOMRIGHT", frame.icons[#frame.icons], "BOTTOMRIGHT", 0, -self.db.char.barwidth)
			else
				frame.bar:SetPoint("BOTTOMLEFT", frame.icons[1], "TOPLEFT")
				frame.bar:SetPoint("TOPRIGHT", frame.icons[#frame.icons], "TOPRIGHT", 0, self.db.char.barwidth)
			end
		else
			if (self.db.char.side == "B") then
				frame.bar:SetPoint("TOPLEFT", frame.icons[1], "TOPRIGHT")
				frame.bar:SetPoint("BOTTOMRIGHT", frame.icons[#frame.icons], "BOTTOMRIGHT", self.db.char.barwidth, 0)
			else
				frame.bar:SetPoint("TOPRIGHT", frame.icons[1], "TOPLEFT")
				frame.bar:SetPoint("BOTTOMLEFT", frame.icons[#frame.icons], "BOTTOMLEFT", -self.db.char.barwidth, 0)
			end
		end
		self:SetFrameSize()
	end
end

-- SetOrientation
function module:SetOrientation(info, n)
	module.db.char.orientation = n

	local frame = self.frame
	if (frame) then
		local prev
		for i,icon in ipairs(frame.icons) do
			icon:ClearAllPoints()
			if (prev) then
				icon:SetPoint("TOPLEFT", prev, n == "H" and "TOPRIGHT" or "BOTTOMLEFT")
			else
				icon:SetPoint("TOPLEFT")
			end
			prev = icon
		end
		self:SetBarOrientation()
	end
end

-- ToggleActive
function module:ToggleActive()
	self.db.char.enable = not self.db.char.enable
	self:ShowWhen()
end

-- SetIconSize
function module:SetIconSize(info, n)
	self.db.char.size = n
	
	if (self.frame) then
		for i,icon in pairs(self.frame.icons) do
			icon:SetWidth(n)
			icon:SetHeight(n)
		end

		self:SetFrameSize()
		self:SavePosition()
	end
end

-- SetScale
function module:SetScale()
	if (self.frame) then
		self.frame:SetScale(self.db.char.scale)
	end
end

-- ShowWhen
function module:ShowWhen()
	if (self.db.char.enable) then
		if (not self.frame) then
			self:InitFrames()
		end

		local show = self.db.char.always
		if (not show) then
			if (UnitCanAttack("player", "target")) then
				local pvpInstance = select(2, IsInInstance()) == "pvp"
				if (self.db.char.showRaid) then
					show = GetNumRaidMembers() > 0 and not pvpInstance
				elseif (self.db.char.showParty) then
					show = GetNumPartyMembers() > 0 and not pvpInstance
				elseif (self.db.char.showSolo) then
					show = GetNumRaidMembers() == 0 and GetNumPartyMembers() == 0
				end
			end
		end

		if (show) then
			self.frame:Show()
		else
			self.frame:Hide()
		end
	elseif (self.frame) then
		self.frame:Hide()
	end
end

-- PLAYER_TARGET_CHANGED
function module:PLAYER_TARGET_CHANGED()
	self:ShowWhen()
end

-- frameOnUpdate
local function frameOnUpdate(self, elapsed)
	if (module.rosterUpdate) then
		module:OnRaidRosterUpdate()
	end
	module.frame:SetScript("OnUpdate", nil)
end

-- PARTY_MEMBERS_CHANGED
function module:PARTY_MEMBERS_CHANGED()
	self.rosterUpdate = true
	self.frame:SetScript("OnUpdate", frameOnUpdate)
end

-- OnRaidRosterUpdate
function module:OnRaidRosterUpdate()
	self.distribution = nil
	if (select(2, IsInInstance()) ~= "pvp") then
		if (GetNumRaidMembers() > 0) then
			self.distribution = "RAID"
		elseif (GetNumPartyMembers() > 0) then
			self.distribution = "PARTY"
		end
	end

	if (self.distribution) then
		self:SendCommMessage(commPrefix, "HELLO "..self.version, self.distribution)
	end

	self:ScanRoster()
	if (self.distribution or self.db.char.always) then
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
	else
		self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	end

	self:PLAYER_TARGET_CHANGED()
end

do
	local function iter(t)
		local key = t.id
		if (t.mode == "raid") then
			if (key > t.r) then
				del(t)
				return nil
			end
			ret = "raid"..key
		else
			if (key > t.p) then
				del(t)
				return nil
			end
			ret = key == 0 and "player" or "party"..key
		end
		t.id = key + 1
		return ret
	end

	-- IterateRoster
	function module:IterateRoster()
		local t = new()
		if (GetNumRaidMembers() > 0) then
			t.mode = "raid"
			t.id = 1
			t.r = GetNumRaidMembers()
		else
			t.mode = "party"
			t.id = 0
			t.p = GetNumPartyMembers()
		end
		return iter, t
	end
end

-- ScanRoster
function module:ScanRoster()
	del(self.potential)
	self.potential = new()
	del(self.spellPotential)
	self.spellPotential = new()

	if (self.talents) then
		local oldTalents = new()
		for name,talents in pairs(self.talents) do
			oldTalents[name] = true
		end
	end

	self.classcount = setmetatable({}, {__index = function() return 0 end})
	for unit in self:IterateRoster() do
		if (oldTalents) then
			local name = UnitFullName(unit)
			oldTalents[name] = nil
		end
		
		local _, class = UnitClass(unit)
		if (class) then
			self.classcount[class] = self.classcount[class] + 1

			local talents = self:GetTalents(unit)

			local c = self.classes[class]
			if (c) then
				for key, name in pairs(c) do
					local info = self.lookup[name]

					if (not info.requiredTalent) then
						self.potential[key] = true
						self.spellPotential[name] = true
					else
						if (talents) then
							local level = self:HasTalent(unit, info.requiredTalent)
							if (level and (level == info.maxTalentPoints or not info.maxTalentPoints)) then
								self.potential[key] = true
								self.spellPotential[name] = true
							end
						end
					end
				end
			end
		end
	end

	if (oldTalents) then
		for name in pairs(oldTalents) do
			self.talents[name] = del(self.talents[name])
		end
		del(oldTalents)
	end

	if (self.frame) then
		self.frame:UpdateBuffs()
	end
end

-- HasTalent
function module:HasTalent(unit, required)
	if (self.classTalentData) then
		local _, class = UnitClass(unit)
		if (class) then
			local data = self.classTalentData[class]
			if (not data) then
				if (class == select(2,UnitClass("player"))) then
					self:GetClassTalentData("player")
				elseif ((UnitInRaid(unit) or UnitInParty(unit)) and UnitIsConnected(unit)) then
					if (UnitIsVisible(unit) and CanInspect(unit)) then
						if (not UnitIsPlayer(unit)) then
							NotifyInspect(unit)
						end
						self:GetClassTalentData(unit)
					end
				end
				data = self.classTalentData[class]
			end

			if (data) then
				local info = data.list[required]
				if (info) then
					local index = info.index
					local tree = info.treeIndex
	
					local talents = self.talents and self.talents[UnitFullName(unit)]
					if (talents) then
						local str = talents[tree]
						if (str) then
							local amount = tonumber(str:sub(index, index))
							return amount > 0 and amount or nil
						end
					end
				end
			end
		end
	end
end

-- OnReceiveTalents
function module:OnReceiveTalents(fullname, talents)
	if (self.frame) then
		self.frame:UpdateBuffs()
	end
end

-- GetTalents
do
	local RockComm = LibStub("LibRockComm-1.0", true)
	if (RockComm) then
		local function talentReceptor(sender, data)
			if (not module.talents) then
				module.talents = new()
			end

			if (type(data) == "table" and #data == 3 and type(data[1]) == "table") then
				local n = new()
				
				for i = 1,3 do
					local str = ""
					for j = 1,#data[i] do
						str = str .. tostring(data[i][j])
					end
					tinsert(n, str)
				end
				module.talents[sender] = n
				module:OnReceiveTalents(sender, n)
			end
		end
		RockComm:AddTalentReceptor(talentReceptor)
	end
	
	function module:GetTalents(unit)
		local name = UnitFullName(unit)
		local talents = self.talents and self.talents[name]

		if (not talents) then
			if (UnitIsUnit("player", unit)) then
				talents = self:TalentQuery_Ready(nil, UnitName(unit))
			elseif ((UnitInRaid(unit) or UnitInParty(unit)) and UnitIsConnected(unit)) then
				if (UnitIsVisible(unit) and CanInspect(unit)) then
					TalentQuery:Query(unit)

				elseif (self.users and self.users[name]) then
					self:SendCommMessage(commPrefix, "REQUESTTALENTS", "WHISPER", name)

				elseif (RockComm) then
					RockComm:QueryTalents("WHISPER", name)
				end
			end
		end

		return talents
	end
end

-- GetClassTalentData
function module:GetClassTalentData(unit)
	local _, class = UnitClass(unit)

	if (class) then
		local data = self.classTalentData and self.classTalentData[class]
		if (not data) then
			local isnotplayer = not UnitIsUnit("player", unit)
			if (GetNumTalentTabs(isnotplayer) > 0) then
				data = new()
				data.list = new()

				for tab = 1, GetNumTalentTabs(isnotplayer) do
					local tree = new()
					local pointssptent
					tree.name, tree.icon, pointssptent = GetTalentTabInfo(tab, isnotplayer)
					tinsert(data, tree)

					tree.list = new()
					for i = 1,GetNumTalents(tab, isnotplayer) do
						local name, icon, tier, column, currentRank, maxRank, isExceptional, meetsPrereq = GetTalentInfo(tab, i, isnotplayer)

						local entry = new()
						entry.name = name
						entry.icon = icon
						entry.tier = tier
						entry.column = column
						entry.maxRank = maxRank
						entry.index = i
						entry.treeIndex = tab
						tinsert(tree.list, entry)
						data.list[name] = entry
					end
				end

				if (next(data)) then
					if (not self.classTalentData) then
						self.classTalentData = new()
					end
					self.classTalentData[class] = data
				end
			end
		end
	end
end

-- TalentQuery_Ready
function module:TalentQuery_Ready(e, name, realm)
	local namerealm = realm and realm ~= "" and name .. "-" .. realm or name
	local isnotplayer = (namerealm ~= UnitName("player"))

	self:GetClassTalentData(namerealm)

	if (GetNumTalentTabs(isnotplayer) and GetNumTalentTabs(isnotplayer) > 0) then
		local n = new()
		for tab = 1, GetNumTalentTabs(isnotplayer) do
			local str
			for i = 1,GetNumTalents(tab, isnotplayer) do
				local name, icon, tier, column, currentRank, maxRank, isExceptional, meetsPrereq = GetTalentInfo(tab, i, isnotplayer)
				str = (str or "") .. tostring(currentRank)
			end
			tinsert(n, str)
		end

		if (not self.talents) then
			self.talents = new()
		end
		self.talents[namerealm] = n
		self:OnReceiveTalents(namerealm, n)

		return n
	end
end

-- OnCommReceived
function module:OnCommReceived(prefix, msg, channel, sender)
	if (sender == playerName) then
		return
	end

	if (prefix == commPrefix) then
		local cmd, str = msg:match("^(%a+) *(.*)$")
		if (not cmd) then
			return
		end

		if (cmd == "TALENTS") then
			local list = new(strsplit(",", str))
			if (next(list)) then
				self.talents[sender] = list
				self:OnReceiveTalents(sender, list)
			end
			del(list)

		elseif (cmd == "REQUESTTALENTS") then
			local t = self.talents[UnitName("player")]
			if (t) then
				self:SendCommMessage(commPrefix, format("TALENTS %s,%s,%s", t[1], t[2], t[3]), "WHISPER", sender)
			end

		elseif (cmd == "HELLO") then
			if (channel ~= "WHISPER" and sender ~= UnitName("player")) then
				if (not self.users) then
					self.users = {}
				end
				self.users[sender] = tonumber(str) or str
				if (str ~= "ZOMG") then
					self:SendCommMessage(commPrefix, "HELLO "..self.version, "WHISPER", sender)
				end
			end
		end
	end
end

-- OpenConfig
function module:OpenConfig(input)
	if (input == "test") then
		for class,data in pairs(self.classTalentData) do
			if (not RAID_CLASS_COLORS[class]) then
				self:Print(format("Bad class '%s' in classTalentData", tostring(class)))
			elseif (not data.list or not data[1] or not data[2] or not data[3]) then
				self:Print(format("Bad data for '%s' in classTalentData", tostring(class)))
			end
		end
		for name,data in pairs(self.talents) do
			if (type(data) ~= "table") then
				self:Print(format("Invalid talent info for %s", tostring(name)))
			elseif (#data < 3) then
				self:Print(format("Incomplete talent info for %s", tostring(name)))
			elseif (type(data[1]) ~= "string" or type(data[2]) ~= "string" or type(data[3]) ~= "string") then
				self:Print(format("Talent info is not strings for %s", tostring(name)))
			end
		end

	elseif (input == "config" or not InterfaceOptionsFrame:IsResizable()) then
		module.options.args.aceConfig.guiHidden = true
		InterfaceOptionsFrame:Hide()
		AceConfigDialog:SetDefaultSize("Utopia", 500, 550)
		AceConfigDialog:Open("Utopia")
	else
		module.options.args.aceConfig.guiHidden = false
		InterfaceOptionsFrame_OpenToCategory(self.optFrame)
	end
end

-- OnModuleInitialize
function module:OnInitialize()
	new, del, copy, deepDel = self.new, self.del, self.copy, self.deepDel
	Gradient, UnitFullName, ClassColour, SmoothColour = self.Gradient, self.UnitFullName, self.ClassColour, self.SmoothColour
	rotate = Utopia.rotate

	module.options.name = Gradient(L["TITLE"], 1, 0.1, 0.1, 1, 1, 0.1)

	TalentQuery.RegisterCallback(self, "TalentQuery_Ready")

	playerClass = select(2, UnitClass("player"))

	self.db = LibStub("AceDB-3.0"):New("UtopiaDB", defaults, "Default")

	self.db:RegisterDefaults({
		char = {
			activeColour = {r = 1, g = 1, b = 1, a = 1},
			inactiveColour = {r = 1, g = 0, b = 0, a = 1},
			unavailableColour = {r = 1, g = 1, b = 1, a = 0.25},
			enable = true,
			size = 20,
			scale = 1,
			texture = "Blizzard",
			showSolo = false,
			showParty = true,
			showRaid = true,
			orientation = "H",
			side = "B",
			barwidth = 7,
		}
	})

	self.optFrame = AceConfigDialog:AddToBlizOptions("Utopia", "Utopia")

	module.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self:RegisterChatCommand("utopia", "OpenConfig")
	self:RegisterComm(commPrefix)

	self.OnInitialize = nil
end

-- OnModuleEnable
function module:OnEnable()
	if (self.db) then
		local class = select(2, UnitClass("player"))
		if (class ~= playerClass and self.OnModuleInitialize) then
			self:OnModuleInitialize()
		end

		self:InitData()
		self:OnRaidRosterUpdate()
		self:PLAYER_TARGET_CHANGED()

		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED")
	end
end

-- OnDisable
function module:OnDisable()
	self.lookup = del(self.lookup)
	self.frame:Hide()
end
